#include <windows.h>

#define KEYHOOK_API __declspec(dllexport)

KEYHOOK_API BOOL APIENTRY kh_hook(HWND hWnd, UINT msg);
KEYHOOK_API BOOL APIENTRY kh_hook_ll(HWND hWnd, UINT msg);
KEYHOOK_API BOOL APIENTRY kh_unhook();

#pragma comment(linker, "/SECTION:.shared,RWS")
#pragma data_seg(".shared")
static HHOOK hHook = NULL;
static HWND hHostWnd = NULL;
static HINSTANCE hInstance = NULL;
static UINT hHostWndMsg = 0;
#pragma data_seg()

static LRESULT CALLBACK KeyboardProc(int nCode, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK LowLevelKeyboardProc(int nCode, WPARAM wParam, LPARAM lParam);

BOOL WINAPI DllMain(HINSTANCE hInstDLL, DWORD fdwReason, LPVOID lpReserved)
{
    if(fdwReason == DLL_PROCESS_ATTACH) 
        hInstance = hInstDLL;
    return TRUE;
}

KEYHOOK_API BOOL APIENTRY kh_hook(HWND hWnd, UINT msg)
{
    if(hHook != NULL && kh_unhook() == FALSE) return FALSE;

    hHook = SetWindowsHookEx(WH_KEYBOARD, KeyboardProc, hInstance, 0);
    if (hHook == NULL) return FALSE;

    hHostWnd = hWnd;
    hHostWndMsg = msg;

    return TRUE;
}

KEYHOOK_API BOOL APIENTRY kh_hook_ll(HWND hWnd, UINT msg)
{
    if(hHook != NULL && kh_unhook() == FALSE) return FALSE;

    hHook = SetWindowsHookEx(WH_KEYBOARD_LL, LowLevelKeyboardProc, hInstance, 0);
    if (hHook == NULL) return FALSE;

    hHostWnd = hWnd;
    hHostWndMsg = msg;

    return TRUE;
}

KEYHOOK_API BOOL APIENTRY kh_unhook()
{
    if(hHook != NULL && UnhookWindowsHookEx(hHook)) {
        hHook = NULL;
        return TRUE;
    }
    return FALSE;
}

LRESULT CALLBACK KeyboardProc(int nCode, WPARAM wParam, LPARAM lParam)
{
    LRESULT ret = CallNextHookEx(hHook, nCode, wParam, lParam);
    if (nCode == HC_ACTION)
        ret |= SendMessage(hHostWnd, hHostWndMsg, wParam, lParam);
    return ret;
}

LRESULT CALLBACK LowLevelKeyboardProc(int nCode, WPARAM wParam, LPARAM lParam)
{
    LRESULT ret = CallNextHookEx(hHook, nCode, wParam, lParam);
    if (nCode == HC_ACTION) {
        LPKBDLLHOOKSTRUCT p = (LPKBDLLHOOKSTRUCT)lParam;
        ret |= SendMessage(hHostWnd, hHostWndMsg, wParam << 16 | p->vkCode, p->flags << 24 | p->scanCode << 16);
    }
    return ret;
}
